"""
    Global engine query helper module
"""
import json
from sdc_helpers import decorators
from sdc_helpers.query_helper import QueryHelper
from sqlalchemy import and_
from sqlalchemy.exc import InternalError
from sqlalchemy.orm import with_expression
from sdc_engine_helpers.models.service import Service
from sdc_engine_helpers.models.subscription import Subscription


class EngineQueryHelper(QueryHelper):
    """
        Reusable generic engine query helper
    """
    # pylint: disable=singleton-comparison, arguments-differ

    @decorators.query_exception_handler(exceptions=(Exception, InternalError,))
    def get_service(
            self,
            *,
            slug: str,
            for_update: bool = False
    ):
        """
            Get the requested service from the database with the maintenance_in_progress property

            args:
                service_slug (int): The requested service slug

            returns:
                service (Service): The Service model
        """
        if not for_update:
            return super().get_service(slug=slug)

        return self.get_query(Service, for_update).filter(
            and_(
                Service.slug == slug,
                Service.deleted_at == None
            )
        ).options(
            with_expression(
                Service.maintenance_in_progress,
                Service.maintenance_in_progress_expression()
            )
        ).first()

    @decorators.query_exception_handler(exceptions=(Exception, InternalError,))
    def get_engine_entity(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            entity_key: str,
            from_cache: bool = True,
            for_update: bool = False
    ):
        """
            Get a specific entity from a specific engine in a subscription

            args:
                client_id (int): The client id of the subscription
                service_id (int): The service id of the subscription
                engine_slug (str): Subscription property engine
                entity_key (str): The key of the entity in the engine
                from_cache (bool): Retrieve the entities from cache - Default True
                for_update (bool): Reserve the subscription for exclusive use until
                                   commit or session is closed - Default False

            returns:
                entity_list (list): A list of the requested entity in the specified engine
                engine_index (int): Index of the engine index in properties (None if from cache)
        """
        entities_redis_key = (
            'subscription-{client_id}-{service_id}-{entity_key}-{engine_slug}'.format(
                client_id=client_id,
                service_id=service_id,
                entity_key=entity_key,
                engine_slug=engine_slug
            )
        )
        entities_redis = self.redis_helper.redis_get(key=entities_redis_key)
        engine_index = None

        if (
                not from_cache or
                not entities_redis
        ):
            query = self.get_query(Subscription, for_update).filter(
                and_(
                    Subscription.client_id == client_id,
                    Subscription.service_id == service_id,
                    Subscription.deleted_at == None
                )
            ).options(
                with_expression(
                    Subscription.engine_index,
                    Subscription.engine_index_expression(
                        engine_slug=engine_slug
                    )
                ),
                with_expression(
                    Subscription.engine_node,
                    Subscription.engine_node_expression(
                        engine_slug=engine_slug,
                        entity_key=entity_key
                    )
                )
            )

            subscription = query.first()

            entities = None
            if subscription is not None:
                entities = subscription.engine_node

                if entities:
                    engine_index = int(subscription.engine_index)
                    entities = json.loads(entities)

            self.redis_helper.redis_set(
                key=entities_redis_key,
                value=json.dumps(entities, default=str)
            )
        else:
            entities = json.loads(entities_redis)

        return entities, engine_index

    def get_dataset(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            from_cache: bool = True,
            for_update: bool = False,
            **kwargs
    ):
        """
            Get a specific dataset for an engine in subscription properties

            args:
                client_id (int): The client id of the subscription
                service_id (int): The service id of the subscription
                engine_slug (str): Subscription property engine
                from_cache (bool): Retrieve the datasets from cache - Default True
                for_update (bool): Reserve the subscription for exclusive use until
                                   commit or session is closed - Default False
                kwargs (dict):
                    label (str): The dataset's label

            returns:
                dataset_dict (dict): The requested campaign dictionary, if found
        """
        label = kwargs.get('label', None)

        datasets, _ = self.get_engine_entity(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            entity_key='datasets',
            from_cache=from_cache,
            for_update=for_update
        )

        if datasets is not None:
            for index, dataset in enumerate(datasets):
                if dataset.get('label') == label:
                    return dataset, index

        return None, None

    def get_campaign(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            from_cache: bool = True,
            for_update: bool = False,
            **kwargs
    ):
        """
            Get a specific campaign for an engine in subscription properties

            args:
                client_id (int): The client id of the subscription
                service_id (int): The service id of the subscription
                engine_slug (str): Subscription property engine
                from_cache (bool): Retrieve the campaigns from cache - Default True
                for_update (bool): Reserve the subscription for exclusive use until
                                   commit or session is closed - Default False
                kwargs (dict):
                    arn (str): The campaign's arn
                    recipe (str): The recipe of the campaign e.g related_items
                    event_type (str): Supplied if the campaign only involves certain events e.g
                                      only listing views, but not listing enquiries

            returns:
                campaign_dict (dict): The requested campaign,  if found
        """
        campaigns, _ = self.get_engine_entity(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            entity_key='campaigns',
            from_cache=from_cache,
            for_update=for_update
        )

        if campaigns is not None:
            return self.match_arn_or_recipe_event_type(entities=campaigns, **kwargs)

        return None, None

    def get_solution(
            self,
            *,
            client_id: int,
            service_id: int,
            engine_slug: str,
            from_cache: bool = True,
            for_update: bool = False,
            **kwargs
    ):
        """
            Get a specific solution for an engine in subscription properties

            args:
                client_id (int): The client id of the subscription
                service_id (int): The service id of the subscription
                engine_slug (str): Subscription property engine
                from_cache (bool): Retrieve the solutions from cache - Default True
                for_update (bool): Reserve the subscription for exclusive use until
                                   commit or session is closed - Default False
                kwargs (dict):
                    arn (str): The campaign's arn
                    recipe (str): The recipe of the solution e.g related_items
                    event_type (str): Supplied if the solution only involves certain events e.g
                                      only listing views, but not listing enquiries

            returns:
                solution_dict (dict): The requested solution dictionary, if found
        """
        solutions, _ = self.get_engine_entity(
            client_id=client_id,
            service_id=service_id,
            engine_slug=engine_slug,
            entity_key='solutions',
            from_cache=from_cache,
            for_update=for_update
        )

        if solutions is not None:
            return self.match_arn_or_recipe_event_type(entities=solutions, **kwargs)

        return None, None

    @staticmethod
    def match_arn_or_recipe_event_type(*, entities: list, **kwargs):
        """
            Match a entity on arn or recipe/event_type

            args:
                entities (list): List of entities being searched
                kwargs (dict):
                    arn (str): The entity's arn
                    recipe (str): The recipe of the entity e.g related_items
                    event_type (str): Supplied if the entity only involves certain events e.g
                                      only listing views, but not listing enquiries

            returns:
                entity_dict (dict/None): The requested entity dictionary, if found
        """
        arn = kwargs.get('arn', None)
        recipe = kwargs.get('recipe', None)
        event_type = kwargs.get('event_type', None)

        for index, entity in enumerate(entities):
            if (
                    (
                        arn is not None and entity.get('arn') == arn
                    ) or
                    (
                        entity.get('recipe') == recipe and
                        (event_type is None or entity.get('event_type') == event_type)
                    )
            ):
                return entity, index

        return None, None

    @staticmethod
    def update_engine_entity(
            *,
            subscription: Subscription,
            engine_index: int,
            entity_key: str,
            value
    ):
        """
            Update a specific entity in a specific engine in a subscription

            args:
                subscription (Subscription): The subscription model to update
                engine_index (str): The engine index in the Subscription model properties
                entity_key (str): The key of the entity in the engine
                value: The data to set the entity key node to
        """
        subscription.properties['engines'][engine_index][entity_key] = value

    def update_datasets(
            self,
            *,
            subscription: Subscription,
            engine_index: int,
            value
    ):
        """
            Update datasets in a specific engine in a subscription

            args:
                subscription (Subscription): The subscription model to update
                engine_index (str): The engine index in the Subscription model properties
                value: The data to set the entity key node to
        """
        self.update_engine_entity(
            subscription=subscription,
            engine_index=engine_index,
            entity_key='datasets',
            value=value
        )

    def update_campaigns(
            self,
            *,
            subscription: Subscription,
            engine_index: int,
            value
    ):
        """
            Update campaigns in a specific engine in a subscription

            args:
                subscription (Subscription): The subscription model to update
                engine_index (str): The engine index in the Subscription model properties
                value: The data to set the entity key node to
        """
        self.update_engine_entity(
            subscription=subscription,
            engine_index=engine_index,
            entity_key='campaigns',
            value=value
        )

    def update_solutions(
            self,
            *,
            subscription: Subscription,
            engine_index: int,
            value
    ):
        """
            Update solutions in a specific engine in a subscription

            args:
                subscription (Subscription): The subscription model to update
                engine_index (str): The engine index in the Subscription model properties
                value: The data to set the entity key node to
        """
        self.update_engine_entity(
            subscription=subscription,
            engine_index=engine_index,
            entity_key='solutions',
            value=value
        )

    def flush_subscription_properties_cache(self, subscription: Subscription):
        """
            Match a entity on arn or recipe/event_type

            args:
                subscription (Subscription): The subscription object to flush cache for
        """
        subscription_redis_key_pattern = (
            'subscription-{client_id}-{service_id}*'.format(
                client_id=subscription.client_id,
                service_id=subscription.service_id
            )
        )
        keys = self.redis_helper.redis_keys(
            pattern=subscription_redis_key_pattern
        )

        if len(keys) > 0:
            self.redis_helper.redis_delete(keys=keys)



"""
   SDC Engine subscription model module
"""
from sqlalchemy import func, case, literal
from sqlalchemy.orm import query_expression
from sqlalchemy.sql.functions import _FunctionGenerator
from sdc_helpers.models.subscription import Subscription as BaseSubscription


class Subscription(BaseSubscription):
    """
       SDC Engine Subscription model class
    """
    # pylint: disable=too-few-public-methods
    extend_existing = True

    engine_index = query_expression()
    engine_node = query_expression()

    @classmethod
    def engine_json_search_expression(cls, engine_slug: str) -> _FunctionGenerator:
        """
            JSON_SEARCH SQL function to find the path of a specific engine_slug
            in subscription properties

            args:
                engine_slug (str): The engine slug to search for

            returns:
                function (_FunctionGenerator): The JSON_SEARCH SQL function

        """
        return func.json_search(
            cls.properties,
            'one',
            engine_slug,
            None,
            '$.engines[*].slug'
        )

    @classmethod
    def engine_index_expression(cls, engine_slug: str):
        """
            SQL expression to find the index of an engine slug in properties

            args:
                engine_slug (str): The engine slug to search for

            returns:
                function (_FunctionGenerator): The JSON_SEARCH SQL function

            Example SQL expression:

                REGEXP_REPLACE(
                    JSON_SEARCH(properties, "one", "personalize", NULL, "$.engines[*].slug"),
                    '[^0-9]',
                    ''
                )
        """
        return func.regexp_replace(
            cls.engine_json_search_expression(engine_slug=engine_slug),
            '[^0-9]',
            ''
        )

    @classmethod
    def engine_node_expression(cls, engine_slug: str, entity_key: str) -> _FunctionGenerator:
        """
            SQL expression to get an entity node for an engine slug in properties

            Get the JSON path of the given engine_slug e.g "$.engines[0].slug

            If the the JSON search did not return a path return None
            Else get the path of the entity key being looked for by just replacing
            `slug` by `{entity_key}` e.g "$.engines[0].slug" -> "$.engines[0].campaigns"

            args:
                engine_slug (str): The engine slug to search for
                entity_key (str): The entity key to search for

            returns:
                function (_FunctionGenerator): The JSON_SEARCH SQL function

            Example SQL expression:

                CASE
                    WHEN JSON_SEARCH(
                        properties,
                        "one",
                        "personalize",
                        NULL,
                        "$.engines[*].slug"
                    ) IS NULL THEN NULL
                    ELSE JSON_EXTRACT(
                        properties,
                        JSON_UNQUOTE(
                            REPLACE(
                                JSON_SEARCH(
                                    properties,
                                    "one",
                                    "personalize",
                                    NULL,
                                    "$.engines[*].slug"
                                ),
                                'slug',
                                'campaigns'
                            )
                        )
                    )
                END
        """
        engine_json_search = cls.engine_json_search_expression(engine_slug=engine_slug)

        return case(
            [
                (
                    engine_json_search == literal('NULL'),
                    None
                )
            ],
            else_=func.json_extract(
                Subscription.properties,
                func.json_unquote(
                    func.replace(
                        engine_json_search,
                        'slug',
                        entity_key
                    )
                )
            )
        )



"""
    Queue utilities module
"""
import os
import uuid
from datetime import datetime
import json
import boto3
from sdc_helpers import utils
from sdc_helpers.models.client import Client


def queue_s3_store(
        *,
        content: str,
        file_path: str,
):
    """
       Queue storage into S3

           args:
               content (str): The file content of the file in S3
               file_path (str): The S3 file path
    """
    sqs = boto3.client('sqs')
    queue_url = os.getenv('QUEUE_URL')
    if queue_url is None:
        raise Exception('Please set up the QUEUE_URL environment variable')
    s3_storer_lambda = os.getenv('S3_STORER_LAMBDA')
    if s3_storer_lambda is None:
        raise Exception('Please set up the S3_STORER_LAMBDA environment variable')

    sqs_body = {
        'bucket': 'base-lambda-requests',
        'file_path': file_path,
        'content': content
    }

    sqs_message_attributes = {
        'lambda_function': {
            'DataType': 'String',
            'StringValue': s3_storer_lambda
        }
    }

    sqs.send_message(
        QueueUrl=queue_url,
        MessageBody=json.dumps(sqs_body, indent=2),
        MessageAttributes=sqs_message_attributes
    )


def queue_s3_recommendations_store(
        *,
        client: Client,
        results: list,
        item_id: int,
        context: dict
):
    """
        Queue recommendation results storage into S3

            args:
                client (Client): The client model involved
                results (list): A results of the recommendation
                item_id (int): The item id of the requested recommendation
                context (dict): Additional information of the request
    """
    file_path = '{client}/recommendations/{uuid}.txt'.format(
        client=client.name,
        uuid=str(uuid.uuid4())
    )

    domain = utils.dict_query(
        dictionary=context,
        path='domain'
    )

    session_hash = utils.dict_query(
        dictionary=context,
        path='sessionHash'
    )

    content = {
        'domain': domain,
        'item_id': item_id,
        'session_hash': session_hash,
        'results': results,
        'timestamp': datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S')
    }

    queue_s3_store(
        content=json.dumps(content, indent=2),
        file_path=file_path
    )



"""
    Solution state manager module
        - implements Workflows that call AWS Vendor API's.
"""
from datetime import datetime
import sdc_helpers.utils as utils
import boto3
from sdc_engine_helpers.sagemaker.utils import name_from_base
from sdc_engine_helpers.date_utils import DateUtils

class SolutionStateManager:
    """
        Given a database solution object,

        1. get_state
            - call vendor API's to check if solution needs to be updated
            OR
            - check if the solution status in the database needs to be refreshed.
        2. update_state
            - call vendor API's to execute update workflow for solution
            - update status of solution in database

    """
    date_utils = DateUtils()
    # slack_helper = SlackHelper()
    # set path to entities in vendor objects
    solution_status_path = 'TrainingJobStatus'
    # setup available status to look for
    job_statuses = {
        'inProgress':['InProgress'],
        'running': ['InProgress', 'Stopping'],
        'failed':['Failed'],
        'blocking':['InProgress', 'Stopping', 'Failed'],
        'stopped':['Stopped'],
        'success':['Completed', None, 'None']
    }

    required_new_job_fields = [
        "TrainingJobName",
        'AlgorithmSpecification',
        'HyperParameters',
        'InputDataConfig',
        'OutputDataConfig',
        'ResourceConfig',
        'EnableManagedSpotTraining',
        'RoleArn',
        'StoppingCondition'
    ]

    def __init__(self):
        """
            Init a new version of the solution state manager

            Args:
                solution_base_name (str): [description]
        """
        # get vendor api
        self.sagemaker = boto3.client('sagemaker')

    def sagemaker_create_solution(self, training_job_name: str, new_solution_name: str) -> str:
        """Performs a create solution API execution to create new solution version job

            Args:
                training_job_name (str): name of previously successful training job
                new_solution_name (str): new training job name

            Returns:
                (str) : new training job name
        """

        # get current solution
        prev_training_job = self.sagemaker.describe_training_job(
            TrainingJobName=training_job_name
        )

        new_training_job = {}
        for key_, values_ in prev_training_job.items():
            if key_ in self.required_new_job_fields:
                new_training_job.update({
                    key_ : values_
                })

        # update fields
        # generate a unique name with timestamp
        new_training_job_name = name_from_base(
            new_solution_name,
            max_length=63,
            short=False
        )
        new_training_job.update({
            'TrainingJobName': new_training_job_name
        })

        self.sagemaker.create_training_job(**new_training_job)

        return new_training_job['TrainingJobName']

    def get_vendor_version_status(self, *, vendor_solution: dict) -> str:
        """
            Get vendor solution version status

            args:
                vendor_solution (dict): Solution from vendor

            returns:
                status (str): vendor solution version status

        """

        return utils.dict_query(
            dictionary=vendor_solution,
            path=self.solution_status_path
        )

    def get_state(self, *, solution: dict) -> dict:
        """
            Get the current state of the solution from vendor API's.

            Args:
                solution (dict): database solution config
            Return:
                (dict): current state config

        """
        state = {
            'should_create_version': False,
            'should_refresh_version_status': False
        }

        # get vendor solution
        vendor_solution = self.sagemaker.describe_training_job(
            TrainingJobName=solution.get('arn')
        )
        # check if new solution should be updated
        state['should_create_version'] = self.should_create_solution_version(
            solution=solution,
            vendor_solution=vendor_solution
        )

        if state['should_create_version']:
            # update next creation time
            this_version_creation_time = solution.get('next_version_creation_time', None)
            version_creation_frequency = solution.get('version_creation_frequency', None)

            if version_creation_frequency:
                state['next_version_creation_time'] = self.date_utils.get_next_time(
                    this_time=this_version_creation_time,
                    frequency=version_creation_frequency
                )
        else:
            # Check if the database version status needs to be updated
            version_status = solution.get('version_status', None)
            vendor_solution_status = self.get_vendor_version_status(
                vendor_solution=vendor_solution
            )

            # only update database status if refresh is true
            if vendor_solution_status != version_status:
                state['should_refresh_version_status'] = True
                state['new_version_status'] = vendor_solution_status

        return state

    def should_create_solution_version(
            self, *,
            solution: dict,
            vendor_solution: dict
    ) -> bool:
        """
            Check whether a new solution version should be updated

            args:
                solution (dict): Solution from the database
                vendor_solution (dict): Solution from vendor

            returns:
                result (bool): Should create solution version

        """
        # If the latest solution version is still creating in Personalize, don't create
        vendor_solution_version_status = self.get_vendor_version_status(
            vendor_solution=vendor_solution
        )

        if vendor_solution_version_status in self.job_statuses['blocking']:
            return False

        # If solution create job == failed, do not update
        if vendor_solution_version_status in self.job_statuses['failed']:
            # FUTURE: trigger/log an issue
            return False

        version_creation_frequency = solution.get('version_creation_frequency', None)

        if version_creation_frequency:
            # check if it's time to update solution
            next_version_creation_time = solution.get('next_version_creation_time')
            if next_version_creation_time:
                if (
                        datetime.now() <
                        datetime.strptime(next_version_creation_time, "%Y-%m-%d %H:%M:%S")
                ):
                    return False

        return True

    def update_state(self, solution: dict, solution_base_name: str):
        """
            Run the update workflow for solution

            Args:
                solution (dict): database solution config
                solution_base_name (str): base name for new solution
            Return:
                (dict): new solution config

        """
        new_solution = {
            'arn': None,
            'version_status': None,
            'version_last_created_at': None,
            'next_version_creation_time': None
        }

        new_solution_arn = self.sagemaker_create_solution(
            training_job_name=solution['arn'],
            new_solution_name=solution_base_name
        )
        # update the arn with the new arn
        new_solution.update({
            'arn': new_solution_arn
        })

        vendor_solution = self.sagemaker.describe_training_job(
            TrainingJobName=new_solution_arn
        )

        vendor_solution_version_status = self.get_vendor_version_status(
            vendor_solution=vendor_solution
        )
        # default to inProgress
        new_solution.update({
            'version_status': vendor_solution_version_status
        })

        new_solution.update({
            'version_last_created_at': datetime.strftime(
                datetime.now(),
                self.date_utils.get_mysql_date_format()
                )
        })

        version_creation_frequency = solution['version_creation_frequency']

        if version_creation_frequency:
            next_version_creation_time = solution.get('next_version_creation_time', None)
            new_solution.update({
                'next_version_creation_time': self.date_utils.get_next_time(
                    this_time=next_version_creation_time,
                    frequency=version_creation_frequency
                )
            })
        return new_solution



"""
    Date utilities module
"""
from datetime import datetime
from datetime import timedelta

class DateUtils:
    """
        Date utilities class to perform specific date manipulation
    """

    mysql_date_format = '%Y-%m-%d %H:%M:%S'

    def get_next_time(self, *, this_time=None, frequency: int, return_type='str') -> datetime:
        """
            Perform the required maintenance on a list of campaigns

            args:
                this_time: The time to calculate with (str/datetime)
                frequency (int): A period in minutes to calculate next time with
                return_type (str): The requested return type - default string

            returns:
                next_time (datetime): The calculated next time

        """
        if this_time is None:
            this_time = datetime.now()

        if isinstance(this_time, str):
            this_time = datetime.strptime(
                this_time,
                self.mysql_date_format
            )

        next_time = this_time + timedelta(
            minutes=frequency
        )

        # If next time is less than the current time. Set next time to the current time
        # plus the frequency
        now = datetime.now()
        if next_time < now + timedelta(minutes=frequency + 1):
            next_time = now + timedelta(minutes=frequency)

        if return_type == 'str':
            next_time = datetime.strftime(
                next_time,
                self.mysql_date_format
            )

        return next_time

    def get_mysql_date_format(self) -> str:
        """
            Get the defined MySql date format

            returns:
                date_format (str): This class's Mysql date format

        """
        return self.mysql_date_format



"""
    Campaign state manager module
"""
from datetime import datetime
import sdc_helpers.utils as utils
import boto3
from sdc_engine_helpers.date_utils import DateUtils


class CampaignStateManager:
    """
        Given a database campaign object, check whether the Personalize campaign needs to be
        updated OR whether the status in the database needs to be refreshed. If so and the
        original status was CREATE IN_PROGRESS, we need to refresh the campaign's default
        results in the database
    """
    personalize = None
    personalize_runtime = None
    date_utils = None

    def __init__(self):
        self.personalize = boto3.client('personalize')
        self.personalize_runtime = boto3.client('personalize-runtime')
        self.date_utils = DateUtils()

    def get_state(self, *, campaign: dict) -> dict:
        """
            Get the current state of the campaign in Personalize

            args:
                campaign (dict): Campaign from the database

            returns:
                state (dict): Dictionary containing:
                              1) should_update flag
                              2) should_refresh_status
                              3) latest_solution_version_arn
                              4) next_update_time (if frequency specified)
                              5) new_status
                              6) refresh_default_results
                              7) new_default_results

        """
        # Get Personalize campaign from Personalize client, an exception is thrown if
        # the campaign arn is not found
        personalize_campaign_response = self.personalize.describe_campaign(
            campaignArn=campaign.get('arn')
        )
        personalize_campaign = personalize_campaign_response.get('campaign')

        state = {
            'should_update': False,
            'should_refresh_status': False,
            'latest_solution_version_arn': None,
            'next_update_time': None,
            'new_status': None,
            'refresh_default_results': False,
            'new_default_results': None
        }

        personalize_solution = self.get_solution(personalize_campaign=personalize_campaign)

        # Check for update
        state['should_update'] = self.should_update(
            campaign=campaign,
            personalize_campaign=personalize_campaign,
            personalize_solution=personalize_solution
        )

        if state['should_update']:
            state['latest_solution_version_arn'] = utils.dict_query(
                dictionary=personalize_solution,
                path='latestSolutionVersion.solutionVersionArn'
            )

            this_update_time = campaign.get('next_update_time', None)
            update_frequency = campaign.get('update_frequency', None)

            if update_frequency:
                state['next_update_time'] = self.date_utils.get_next_time(
                    this_time=this_update_time,
                    frequency=update_frequency
                )
        else:
            # Check if the database status needs to be updated
            status = campaign.get('status', None)
            personalize_status = self.get_personalize_status(
                personalize_campaign=personalize_campaign
            )

            if personalize_status != status:
                state['should_refresh_status'] = True
                state['new_status'] = personalize_status

                if status == 'CREATE IN_PROGRESS' and personalize_status == 'ACTIVE':
                    personalize_default_results = self.get_default_results(campaign=campaign)
                    default_results = campaign.get('default_results', [])
                    if default_results != personalize_default_results:
                        state['refresh_default_results'] = True
                        state['new_default_results'] = personalize_default_results

        return state

    def get_solution(self, *, personalize_campaign: dict) -> dict:
        """
            Obtain the latest solution version for the solution the campaign uses by:
                1) Describing the solution version the campaign is currently using
                2) Describing the solution for the above solution version

            args:
                personalize_campaign(dict): Campaign from Personalize

            returns:
                result (bool): The latest solution version for the campaign

        """
        personalize_solution_version_response = self.personalize.describe_solution_version(
            solutionVersionArn=personalize_campaign['solutionVersionArn']
        )

        personalize_solution_version = personalize_solution_version_response.get('solutionVersion')

        personalize_solution_response = self.personalize.describe_solution(
            solutionArn=personalize_solution_version['solutionArn']
        )

        return personalize_solution_response.get('solution')

    def should_update(
            self,
            *,
            campaign: dict,
            personalize_campaign: dict,
            personalize_solution: dict
    ) -> bool:
        """
            Check whether a campaign should be updated when all of the following are true

            1) The latest campaign update in Personalize is ACTIVE or FAILED
            2) The campaign isn't on the latest solution version
            3) The latest solution version is ACTIVE
            4) The current time is >= the scheduled next campaign update time (if specified)

            args:
                campaign (dict): Campaign from the database
                personalize_campaign(dict): Campaign from Personalize

            returns:
                result (bool): Should update campaign

        """
        # If the latest solution version is still creating in Personalize, don't update
        personalize_campaign_status = self.get_personalize_status(
            personalize_campaign=personalize_campaign
        )

        if personalize_campaign_status not in ['ACTIVE', 'CREATE FAILED']:
            return False

        # If the Personalize campaign update failed, update the campaign immediately
        if personalize_campaign_status == 'CREATE FAILED':
            return True

        latest_personalize_solution_version_arn = utils.dict_query(
            dictionary=personalize_solution,
            path='latestSolutionVersion.solutionVersionArn'
        )

        # Don't update campaign if its on the latest solution version
        if (
                personalize_campaign.get('solutionVersionArn') ==
                latest_personalize_solution_version_arn
        ):
            return False

        latest_personalize_solution_version_status = utils.dict_query(
            dictionary=personalize_solution,
            path='latestSolutionVersion.status'
        )

        # Don't update campaign if the latest solution version is not active
        if latest_personalize_solution_version_status != 'ACTIVE':
            return False

        next_update_time = campaign.get('next_update_time', None)
        update_frequency = campaign.get('update_frequency', None)

        if next_update_time and update_frequency:
            next_update_time = datetime.strptime(next_update_time, "%Y-%m-%d %H:%M:%S")
            now = datetime.now()

            if now < next_update_time:
                return False

        return True

    @staticmethod
    def get_personalize_status(*, personalize_campaign: dict) -> str:
        """
            Get Personalize status

            args:
                personalize_campaign (dict): Campaign from Personalize

            returns:
                status (str): Personalize status

        """

        return utils.dict_query(
            dictionary=personalize_campaign,
            path='latestCampaignUpdate.status'
        )

    def get_default_results(self, *, campaign: dict) -> list:
        """
            Get Personalize campaign default results

            args:
                campaign (dict): Campaign from the database

            returns:
                default_results (list): Personalize campaign default results

        """
        results = []

        response = self.personalize_runtime.get_recommendations(
            campaignArn=campaign['arn'],
            itemId='notinmodel'
        )

        item_list = response.get('itemList', None)
        if item_list is not None:
            for item in item_list:
                item_id = item.get('itemId', None)
                if item_id is not None:
                    results.append(item_id)

        return results



"""
    Maintenance manager module
"""
from datetime import datetime
from sdc_helpers.models.client import Client
import boto3
from sqlalchemy.exc import SQLAlchemyError
from sdc_engine_helpers import decorators
from sdc_engine_helpers.date_utils import DateUtils
from sdc_engine_helpers.personalize.maintenance.campaign_state_manager import CampaignStateManager
from sdc_engine_helpers.personalize.maintenance.solution_state_manager import SolutionStateManager
from sdc_engine_helpers.personalize.manager import Manager


class MaintenanceManager(Manager):
    """
        Manage AWS Personalize solutions and campaigns
    """
    date_utils = None
    lambda_client = None
    personalize = None
    solution_state_manager = None
    campaign_state_manager = None

    def __init__(self, **kwargs):
        self.service_for_update = True
        super().__init__(**kwargs)
        self.date_utils = DateUtils()
        self.solution_state_manager = SolutionStateManager()
        self.campaign_state_manager = CampaignStateManager()
        self.personalize = boto3.client('personalize')
        process = self.query_helper.get_process(slug='personalize-maintenance')
        if process is not None:
            self.query_helper.process_id = process.id

    @decorators.retry_handler(SQLAlchemyError, total_tries=3, initial_wait=2)
    def perform_maintenance(self):
        """
            Create new Personalize solution versions and update Personalize campaigns
            when required

            Return:
                (bool) : the outcome of maintenance "updated", "not_updated", "blocked"
        """
        maintenance_outcome = "not_updated"

        # check if maintenance can be performed
        self.can_perform_maintenance = self.check_maintenance_is_allowed()

        if self.can_perform_maintenance:
            # Get all active client subscriptions for the supplied service
            for client in self.query_helper.get_clients():
                # Set up a flag to decide whether its necessary to update this subscription
                # after maintenance of this subscription
                update = False

                maintenance_mode, _ = self.query_helper.get_engine_entity(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=self.engine_slug,
                    entity_key='maintenance_mode',
                    from_cache=False
                )

                if maintenance_mode is None or maintenance_mode == 'off':
                    continue

                subscription = self.query_helper.get_subscription(
                    client_id=client.id,
                    service_id=self.service.id,
                    from_cache=False,
                    for_update=True
                )

                solutions, engine_index = self.query_helper.get_engine_entity(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=self.engine_slug,
                    entity_key='solutions',
                    from_cache=False
                )

                if solutions is not None:
                    response = self.perform_solution_maintenance(
                        client=client,
                        solutions=solutions,
                        maintenance_mode=maintenance_mode
                    )

                    if response.get('solutions_updated') is True:
                        self.query_helper.update_solutions(
                            subscription=subscription,
                            engine_index=engine_index,
                            value=response.get('solutions')
                        )
                        update = True

                campaigns, engine_index = self.query_helper.get_engine_entity(
                    client_id=client.id,
                    service_id=self.service.id,
                    engine_slug=self.engine_slug,
                    entity_key='campaigns',
                    from_cache=False
                )

                if campaigns is not None:
                    response = self.perform_campaign_maintenance(
                        client=client,
                        campaigns=campaigns,
                        maintenance_mode=maintenance_mode
                    )

                    if response.get('campaigns_updated') is True:
                        self.query_helper.update_campaigns(
                            subscription=subscription,
                            engine_index=engine_index,
                            value=response.get('campaigns')
                        )
                        update = True

                if update:
                    self.query_helper.update(model=subscription)
                    self.query_helper.flush_subscription_properties_cache(subscription=subscription)
                    maintenance_outcome = "updated"
        else:
            maintenance_outcome = "blocked"

        return maintenance_outcome

    def perform_solution_maintenance(
            self,
            *,
            client: Client,
            solutions: list,
            maintenance_mode: str
    ) -> dict:
        """
            Perform the required maintenance on a list of solutions

            args:
                client (Client): The client this maintenance is being performed for
                solutions (list): A list of solutions from the database to perform maintenance on
                maintenance_mode (str): The maintenance mode of the subscription
                                        (full/database_only)

            returns:
                result (dict): A dictionary with an updated flag and resultant solutions

        """
        # Set up a flag to return if the solutions have been updated
        solutions_updated = False

        for index, solution in enumerate(solutions):
            # The solution arn is required to continue maintenance for this solution
            arn = solution.get('arn', None)

            if arn is None:
                continue

            state = self.solution_state_manager.get_state(
                solution=solution
            )

            if maintenance_mode == 'full' and state.get('should_create_version') is True:
                print(
                    'Building solution version for client: {client_name}'
                    .format(client_name=client.name)
                )

                response = self.personalize.create_solution_version(
                    solutionArn=arn
                )

                print(
                    'Successfully started creation of solution version: {solution_version_arn}'
                    .format(solution_version_arn=response.get('solutionVersionArn'))
                )

                solutions[index]['version_status'] = 'CREATE IN_PROGRESS'
                solutions[index]['version_last_created_at'] = datetime.strftime(
                    datetime.now(),
                    self.date_utils.get_mysql_date_format()
                )

                next_version_creation_time = state.get('next_version_creation_time', None)
                if next_version_creation_time:
                    solutions[index]['next_version_creation_time'] = next_version_creation_time

                solutions_updated = True
            elif state.get('should_refresh_version_status') is True:
                new_version_status = state.get('new_version_status')

                print(
                    'Refreshing version status on solution: {arn} for client: {client_name}'
                    ' to {status}'
                    .format(
                        arn=arn,
                        client_name=client.name,
                        status=new_version_status
                    )
                )

                solutions[index]['version_status'] = new_version_status
                solutions_updated = True

        return {
            'solutions_updated': solutions_updated,
            'solutions': solutions
        }

    def perform_campaign_maintenance(
            self,
            *,
            client: Client,
            campaigns: list,
            maintenance_mode: str
    ) -> dict:
        """
            Perform the required maintenance on a list of campaigns

            args:
                client (Client): The client this maintenance is being performed for
                campaigns (list): A list of campaigns from the database to perform maintenance on
                maintenance_mode (str): The maintenance mode of the subscription
                                        (full/database_only)

            returns:
                result (dict): Dictionary containing:
                               1) campaigns_updated
                               2) Resultant campaigns list

        """
        # Set up a flag to return if the campaigns have been updated
        campaigns_updated = False

        # Some clients share the same campaigns so we return all refreshed defaults results
        # in a dictionary keyed by client and campaign arn. These can be used to replicate to
        # other clients campaigns in the database

        for index, campaign in enumerate(campaigns):
            # The campaign arn is required to continue maintenance for this campaign
            arn = campaign.get('arn', None)

            if arn is None:
                continue

            state = self.campaign_state_manager.get_state(
                campaign=campaign
            )

            if maintenance_mode == 'full' and state.get('should_update') is True:
                latest_solution_version_arn = state.get(
                    'latest_solution_version_arn'
                )
                print(
                    'Updating campaign for client: {client_name} to solution version: {arn}'
                    .format(
                        client_name=client.name,
                        arn=latest_solution_version_arn
                    )
                )

                self.personalize.update_campaign(
                    campaignArn=arn,
                    solutionVersionArn=latest_solution_version_arn
                )

                print('Successfully started updating campaign: {arn}'.format(arn=arn))

                campaigns[index]['last_updated_at'] = datetime.strftime(
                    datetime.now(),
                    self.date_utils.get_mysql_date_format()
                )
                campaigns[index]['status'] = 'CREATE IN_PROGRESS'

                next_update_time = state.get('next_update_time', None)
                if next_update_time is not None:
                    campaigns[index]['next_update_time'] = next_update_time

                campaigns_updated = True
            elif state.get('should_refresh_status') is True:
                new_status = state.get('new_status')

                print(
                    'Refreshing status on campaign: {arn} for client: {client_name} to {status}'
                    .format(
                        arn=arn,
                        client_name=client.name,
                        status=new_status
                    )
                )

                campaigns[index]['status'] = new_status

                if state.get('refresh_default_results') is True:
                    default_results = state.get('new_default_results', None)
                    if default_results is not None:
                        print(
                            'Refreshing default results on campaign: {arn} '
                            'for client: {client_name}'
                            .format(
                                arn=arn,
                                client_name=client.name
                            )
                        )

                        campaigns[index]['default_results'] = default_results

                campaigns_updated = True

        return {
            'campaigns_updated': campaigns_updated,
            'campaigns': campaigns
        }



"""
   SDC engine helper decorators module
"""
import time
from functools import wraps


def retry_handler(
        exceptions,
        total_tries: int = 4,
        initial_wait: float = 0.5,
        backoff_factor: int = 2
):
    """
        Decorator - retry n times when there are exceptions

        args:
            exceptions: Exception instance or list of Exception instances to catch & retry
            total_tries (int): Total retry attempts
            initial_wait (float): initial delay between retry attempts in seconds
            backoff_factor (int): multiplier used to further randomize back off

        return:
            wrapped function's response
    """

    def retry_decorator(func):
        @wraps(func)
        def func_with_retries(*args, **kwargs):
            """
                Wrapper function to decorate function with retry functionality
            """
            tries, delay = 0, initial_wait
            while tries < total_tries:
                try:
                    print(f'Attempt {tries + 1}')
                    return func(*args, **kwargs)
                except exceptions as ex:
                    tries += 1

                    if tries == total_tries:
                        print(
                            'Function: {name}\n'
                            'Failed despite best efforts after {tries} tries'.format(
                                name=func.__name__,
                                tries=tries
                            )
                        )
                        raise ex

                    print(
                        'Function: {name}\n'
                        'Exception {exception}.\n'
                        'Retrying in {delay} seconds!'.format(
                            name=func.__name__,
                            exception=str(ex),
                            delay=delay
                        )
                    )

                    time.sleep(delay)

                    delay *= backoff_factor

        return func_with_retries

    return retry_decorator



"""
    Set up the sdc_engine_helpers package
"""
from setuptools import setup

setup(
    name='sdc-engine-helpers',
    packages=[
        'sdc_engine_helpers',
        'sdc_engine_helpers.models',
        'sdc_engine_helpers.personalize',
        'sdc_engine_helpers.personalize.maintenance',
        'sdc_engine_helpers.personalize.event',
        'sdc_engine_helpers.personalize.recommendations',
        'sdc_engine_helpers.sagemaker',
        'sdc_engine_helpers.sagemaker.recommendations',
        'sdc_engine_helpers.sagemaker.maintenance',
    ],
    install_requires=[
        'sdc-helpers==1.6.1'
    ],
    description='AWS Recommendation Engine Helpers',
    url='http://github.com/RingierIMU/sdc-recommend-engine-helpers',
    version='1.8.2',
    author='Ringier South Africa',
    author_email='tools@ringier.co.za',
    keywords=[
        'pip',
        'helpers',
        'aws',
        'recommendation',
        'personalize',
        'sagemaker'
    ],
    download_url='https://github.com/RingierIMU/sdc-recommend-engine-helpers/archive/v1.8.0.zip'
)
